﻿using FishFactoryContracts.BindingModels;
using FishFactoryContracts.BusinessLogicsContracts;
using FishFactoryContracts.SearchModels;
using FishFactoryContracts.StoragesContracts;
using FishFactoryContracts.ViewModels;
using FishFactoryDataModels.Models;
using Microsoft.Extensions.Logging;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace FishFactoryBusinessLogic
{
    public class ShopLogic : IShopLogic
    {
        private readonly ILogger _logger;
        private readonly IShopStorage _shopStorage;
        public ShopLogic(ILogger<ShopLogic> logger, IShopStorage shopStorage)
        {
            _logger = logger;
            _shopStorage = shopStorage;
        }
        public bool AddCannedInShop(ShopSearchModel model, ICannedModel canned, int count)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (count <= 0)
            {
                throw new ArgumentException("Количество консерв должно быть больше 0", nameof(count));
            }
            _logger.LogInformation("AddCannedInShop. ShopName:{ShopName}.Id:{ Id}", model.ShopName, model.Id);
            var element = _shopStorage.GetElement(model);
            if (element == null)
            {                       
                _logger.LogWarning("AddCannedInShop element not found");
                return false;
            }
            if (element.MaxCountCanneds - element.ListCanneds.Select(x => x.Value.Item2).Sum() < count)
            {
                throw new ArgumentNullException("Магазин переполнен", nameof(count));
            }
            _logger.LogInformation("AddCannedInShop find. Id:{Id}", element.Id);

            if (element.ListCanneds.TryGetValue(canned.Id, out var pair))
            {
                element.ListCanneds[canned.Id] = (canned, count + pair.Item2);
                _logger.LogInformation("AddCannedInShop. Added {count} {canned} to '{ShopName}' shop", count, canned.CannedName, element.ShopName);
            }
            else
            {
                element.ListCanneds[canned.Id] = (canned, count);
                _logger.LogInformation("AddCannedInShop. Added {count} new canned {canned} to '{ShopName}' shop", count, canned.CannedName, element.ShopName);
            }
            _shopStorage.Update(new()
            {
                Id = element.Id,
                Address = element.Address,
                ShopName = element.ShopName,
                DateOpening = element.DateOpening,
                ListCanneds = element.ListCanneds,
                MaxCountCanneds = element.MaxCountCanneds,
            });
            return true;
        }

        public bool Create(ShopBindingModel model)
        {
            CheckModel(model);
            if (_shopStorage.Insert(model) == null)
            {
                _logger.LogWarning("Insert operation failed");
                return false;
            }
            return true;
        }

        public bool Delete(ShopBindingModel model)
        {
            CheckModel(model, false);
            _logger.LogInformation("Delete. Id: {Id}", model.Id);
            if (_shopStorage.Delete(model) == null)
            {
                _logger.LogWarning("Delete operation failed");
                return false;
            }
            return true;
        }

        public ShopViewModel? ReadElement(ShopSearchModel model)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            _logger.LogInformation("ReadElement. ShopName: {ShopName}. Id: {Id}", model.ShopName, model.Id);
            var element = _shopStorage.GetElement(model);
            if (element == null)
            {
                _logger.LogWarning("ReadElement element not found");
                return null;
            }
            _logger.LogInformation("ReadElement find. Id: {Id}", element.Id);
            return element;
        }

        public List<ShopViewModel>? ReadList(ShopSearchModel? model)
        {
            _logger.LogInformation("ReadList. ShopName: {ShopName}. Id: {Id}", model?.ShopName, model?.Id);
            var list = model == null ? _shopStorage.GetFullList() : _shopStorage.GetFilteredList(model);
            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }
            _logger.LogInformation("ReadList. Count: {Count}", list.Count);
            return list;
        }

        public bool Update(ShopBindingModel model)
        {
            CheckModel(model);
            if (_shopStorage.Update(model) == null)
            {
                _logger.LogWarning("Update operation failed");
                return false;
            }
            return true;
        }

        private void CheckModel(ShopBindingModel model, bool withParams = true)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (!withParams)
            {
                return;
            }
            if (string.IsNullOrEmpty(model.ShopName))
            {
                throw new ArgumentNullException("Нет названия магазина", nameof(model.ShopName));
            }
            if (model.MaxCountCanneds < 0)
            {
                throw new ArgumentException("Максимальное количество консерв в магазине не может быть меньше нуля", nameof(model.MaxCountCanneds));
            }
            _logger.LogInformation("Shop. ShopName:{0}.Address:{1}. Id: {2}", model.ShopName, model.Address, model.Id);
            var element = _shopStorage.GetElement(new ShopSearchModel
            {
                ShopName = model.ShopName
            });
            if (element != null && element.Id != model.Id && element.ShopName == model.ShopName)
            {
                throw new InvalidOperationException("Магазин с таким названием уже есть");
            }
        }

        public bool SellCanneds(ICannedModel model, int count)
        {
            return _shopStorage.SellCanneds(model, count);
        }

        public bool AddCanneds(ICannedModel model, int count)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (count <= 0)
            {
                throw new ArgumentException("Количество консерв должно быть больше 0", nameof(count));
            }
            var freeCount = _shopStorage.GetFullList()
                .Select(x => x.MaxCountCanneds - x.ListCanneds
                .Select(p => p.Value.Item2).Sum()).Sum() - count;
            if (freeCount < 0)
            {
                _logger.LogInformation("AddCanned. Не удалось добавить изделия в магазины, они переполнены.");
                return false;
            }
            foreach (var shop in _shopStorage.GetFullList())
            {
                int countFree = shop.MaxCountCanneds - shop.ListCanneds.Select(x => x.Value.Item2).Sum();
                if (countFree < count)
                {
                    if (!AddCannedInShop(new() { Id = shop.Id }, model, countFree))
                    {
                        _logger.LogWarning("AddCannedInShop operation failed.");
                        return false;
                    }
                    count -= countFree;
                }
                else
                {
                    if (!AddCannedInShop(new() { Id = shop.Id }, model, count))
                    {
                        _logger.LogWarning("AddCannedInShop operation failed.");
                        return false;
                    }
                    count = 0;
                }
                if (count == 0)
                {
                    return true;
                }
            }
            return false;
        }
    }
}
